/***********************************************************/
/* Author: Alex Thomson                                    */
/* File: server.c                                          */
/* Purpose: This program acts as an RPSLS server and       */
/*          requires two clients to play the game. The     */
/*          server does all of the processing for the game.*/
/*                                                         */
/* Networking code taken from Molay,                       */
/* "Understanding Unix/Linux Programming"                  */
/* http://wps.prenhall.com/esm_molay_UNIXProg_1/           */
/***********************************************************/

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>

/* Defines */
#define PORTNUM 25000 //The default port of the game
#define HOSTLEN  256 //The default buffer size for all unknown strings
#define TRUE  1 //If you can't figure this one out...
#define FALSE  0 //Then you don't understand this either
#define oops(msg){perror(msg);exit(1);} //Built in error function

//The WHO
#define ROCK 1
#define PAPER 2
#define SCISSORS 3
#define LIZARD 4
#define SPOCK 5

//The ACTION
#define SMASHES 1
#define CUTS 2
#define COVERS 3
#define CRUSHES 4
#define POISONS 5
#define DECAPITATES 6
#define DISPROVES 7
#define VAPORIZES 8
#define EATS 9

//The players
#define PLAYER1 1
#define PLAYER2 2

/* Function Declarations */
int make_server_socket(int );
void do_Connect(char *, int );
void process_rq();
void do_Exit();
void playGame();
void advanceRound();
void determineWinner(int , int );
void checkForGameWin();

/* Global Variable Declaration*/
int intRound = 0; //the round number
char strRound[8]; //the round name
struct game newGame; //the new game struct

//the player struct definition
struct player {
    char name[HOSTLEN]; //the name
    int nameLen; //the length of the name
    int fileDescriptor; // the file descriptor from the player
    FILE *pFile; //the file from the player socket
    int wins ; //the number of wins of the player
};

//the game struct definition
struct game {
    struct player playerA; //the first player
    struct player playerB; //the second player
    int numOfPlayers; //the number of players in the game
};



/***********************************************************/
/* Function: main                                          */
/* Purpose: sets up the server for play to begin           */
/* Parameters: none (void)                                 */
/* Returns: int      exit code                             */
/***********************************************************/
int main(int argc, char **argv){
    int game_Socket; //the socket of the game
    char cmdSize[4]; //the command given

    //clean out the newGame struct
    newGame.numOfPlayers = 0;
    newGame.playerA.wins = 0;
    newGame.playerB.wins = 0;

    printf("Initializing RPSLS Server...\n");

    // Create a socket for the game to run on or exit at failure
    game_Socket = make_server_socket(PORTNUM);
    if ( game_Socket == -1 ) exit(1); // exit if socket creation fails
    printf("Socket created on port %d...\n",PORTNUM);

    //Do forever
    while(TRUE){
        //So long as the number of Players is not 2
        printf("There are %d players in the game.\n",newGame.numOfPlayers);
        while (newGame.numOfPlayers < 2){
            if (newGame.numOfPlayers==0){ //for the first player
                memset(newGame.playerA.name,'\0',HOSTLEN); //clean out the name

                //Set up the file for the player
                newGame.playerA.fileDescriptor = accept(game_Socket,NULL,NULL);
                newGame.playerA.pFile = fdopen(newGame.playerA.fileDescriptor,
                                                "r+" );
                //read out the player's name's size and set the length
                fread (cmdSize,1,3,newGame.playerA.pFile);
                newGame.playerA.nameLen = atoi(cmdSize);

                //read out the name and print diagnostics
                fread (newGame.playerA.name,1,newGame.playerA.nameLen,
                            newGame.playerA.pFile);
                printf("Connected to player 1 : %s\n",newGame.playerA.name);

                //tell the player that he is waiting
                printf("Telling %s to wait for another player.\n",
                        newGame.playerA.name);
                fwrite ("WAIT" , 1,  4 , newGame.playerA.pFile );
                fflush(newGame.playerA.pFile);
                newGame.numOfPlayers++; //increment the number of players
            }
            else if (newGame.numOfPlayers==1){ //if this is the second player
                memset(newGame.playerB.name,'\0',HOSTLEN); //clean out the name

                //Set up the file for the player
                newGame.playerB.fileDescriptor = accept(game_Socket,NULL,NULL);
                newGame.playerB.pFile = fdopen(newGame.playerB.fileDescriptor,
                                                "r+" );
                //read out the player's name's length and set it
                fread (cmdSize,1,3,newGame.playerB.pFile);
                newGame.playerB.nameLen = atoi(cmdSize);

                //read out the name and set it
                fread (newGame.playerB.name,1,newGame.playerB.nameLen,
                        newGame.playerB.pFile);
                printf("Connected to player 2 : %s\n",newGame.playerB.name);

                newGame.numOfPlayers++; //increment the number of players

                if(fork()==0){//fork the game to allow for multiple games
                    playGame(); //play the game
                }
            }
        }
        newGame.numOfPlayers = 0; //reset the number of players in the game
    }

    return 0; //this should never be hit
}

/***********************************************************/
/* Function: playGame                                      */
/* Purpose: begins and manages game play                   */
/* Parameters: none (void)                                 */
/* Returns: none (void)                                    */
/***********************************************************/
void playGame(){
    char oppSendA[HOSTLEN],oppSendB[HOSTLEN]; //filled with the name of the opp

    printf("Telling the players to begin.\n");

    //clean out the memory for the opponent string
    memset(oppSendA,'\0',HOSTLEN);
    memset(oppSendB,'\0',HOSTLEN);

    //format the strings to send the opponent's name to the players
    sprintf(oppSendA,"GETN%03d%s",newGame.playerA.nameLen,newGame.playerA.name);
    sprintf(oppSendB,"GETN%03d%s",newGame.playerB.nameLen,newGame.playerB.name);

    //tell the players their opponent's name
    fwrite (oppSendB, 1, 7 + newGame.playerB.nameLen, newGame.playerA.pFile);
    fflush(newGame.playerA.pFile);
    fwrite (oppSendA, 1, 7 + newGame.playerA.nameLen, newGame.playerB.pFile);
    fflush(newGame.playerB.pFile);

    while(TRUE){ //until the game is over
        advanceRound(); //advance the round

        //send the round number and play start to the players
        fwrite (strRound, 1, 7, newGame.playerA.pFile);
        fflush(newGame.playerA.pFile);
        fwrite (strRound, 1, 7, newGame.playerB.pFile);
        fflush(newGame.playerB.pFile);

        process_rq(); //get ready to process the responses
    }
}

/***********************************************************/
/* Function: process_rq                                    */
/* Purpose: process the responses from the clients         */
/* Parameters: none (void)                                 */
/* Returns: none (void)                                    */
/***********************************************************/
void process_rq() {
    char  cmda[5], arga[4]; //the command and arguments from player 1
    char  cmdb[5], argb[4]; //the command and arguments from player 2

    //Pre-clean the strings
    memset(cmda,'\0',5);memset(arga,'\0',4);
    memset(cmdb,'\0',5);memset(argb,'\0',4);

    //read the command from both players
    fread (cmda,1,4,newGame.playerA.pFile);
    fread (cmdb,1,4,newGame.playerB.pFile);

    if (!strcmp(cmda,"QUIT")){ //if player 1 quits
        printf("Player 1 quit.\n");

        //if player 2 wanted to play, advance his file pointer
        if (!strcmp(cmdb,"PLAY")){fread (argb,1,3,newGame.playerB.pFile);}

        //tell player 2 that he wins by default
        fwrite ("WING" , 1,  4 , newGame.playerB.pFile );
        fflush(newGame.playerB.pFile);

        do_Exit();//exit the server fork
    }
    else if (!strcmp(cmdb,"QUIT")){ //if player 2 quits
        printf("Player 2 quit.\n");

        //if player 1 wants to play, advance his file pointer
        if (!strcmp(cmda,"PLAY")){fread (arga,1,3,newGame.playerA.pFile);};

        //tell player 1 that he wins by default
        fwrite ("WING" , 1,  4 , newGame.playerA.pFile );
        fflush(newGame.playerA.pFile);

        do_Exit();//exit the server fork
    }
    else if ( !strcmp(cmda,"PLAY") && !strcmp(cmdb,"PLAY")){ //if we play
        //read the played values from the players
        fread (arga,1,3,newGame.playerA.pFile);
        fread (argb,1,3,newGame.playerB.pFile);

        printf("Player 1 played %d.\n",atoi(arga));
        printf("Player 2 played %d.\n",atoi(argb));

        determineWinner(atoi(arga),atoi(argb));//determine the round winner
        checkForGameWin(); //check to see if the game is over
    }
}

/***********************************************************/
/* Function: checkForGameWin                               */
/* Purpose: see if the game is over                        */
/* Parameters: none (void)                                 */
/* Returns: none (void)                                    */
/***********************************************************/
void checkForGameWin(){
    if (newGame.playerA.wins>1){ //if player 1 has 2 wins
        //notify the winner and loser
        fwrite ("WING" , 1,  4 , newGame.playerA.pFile );
        fwrite ("LOSG" , 1,  4 , newGame.playerB.pFile );
        fflush(newGame.playerA.pFile);
        fflush(newGame.playerB.pFile);

        do_Exit();//exit the server fork
    }
    else if (newGame.playerB.wins>1){ //if player 2 has 2 wins
        //notify the winner and loser
        fwrite ("LOSG" , 1,  4 , newGame.playerA.pFile );
        fwrite ("WING" , 1,  4 , newGame.playerB.pFile );
        fflush(newGame.playerA.pFile);
        fflush(newGame.playerB.pFile);

        do_Exit();//exit the server fork
    }
}

/***********************************************************/
/* Function: determineWinner                               */
/* Purpose: determines who wins a round based on input     */
/* Parameters: int playA - player A's play                 */
/*             int playB - player B's play                 */
/* Returns: none (void)                                    */
/***********************************************************/
void determineWinner(int playA, int playB){
    int winner = FALSE; //the winner of the round
    int winVal = FALSE; //the winning action
    char printA[8]; //player A's print string
    char printB[8]; //player B's print string

    //Pre-clean the strings
    memset(printA,'\0',8);
    memset(printB,'\0',8);

    //This is the algorithm for determining winners, losers, and the actions
    //It is a large switch case statement.
    switch(playA){
        case ROCK:
            switch(playB){
                case ROCK:      winner = FALSE;     winVal = FALSE;     break;
                case PAPER:     winner = PLAYER2;   winVal = COVERS;    break;
                case SCISSORS:  winner = PLAYER1;   winVal = CRUSHES;   break;
                case LIZARD:    winner = PLAYER1;   winVal = COVERS;    break;
                case SPOCK:     winner = PLAYER2;   winVal = VAPORIZES; break;
            }
        break;
        case PAPER:
            switch(playB){
                case ROCK:      winner = PLAYER1;   winVal = COVERS;    break;
                case PAPER:     winner = FALSE;     winVal = FALSE;     break;
                case SCISSORS:  winner = PLAYER2;   winVal = CUTS;      break;
                case LIZARD:    winner = PLAYER2;   winVal = EATS;      break;
                case SPOCK:     winner = PLAYER1;   winVal = DISPROVES; break;
            }
        break;
        case SCISSORS:
            switch(playB){
                case ROCK:      winner = PLAYER2;   winVal = CRUSHES;   break;
                case PAPER:     winner = PLAYER1;   winVal = CUTS;      break;
                case SCISSORS:  winner = FALSE;     winVal = FALSE;     break;
                case LIZARD:    winner = PLAYER1;   winVal = DECAPITATES;break;
                case SPOCK:     winner = PLAYER2;   winVal = SMASHES;   break;
            }
        break;
        case LIZARD:
            switch(playB){
                case ROCK:      winner = PLAYER2;   winVal = CRUSHES;   break;
                case PAPER:     winner = PLAYER1;   winVal = EATS;      break;
                case SCISSORS:  winner = PLAYER2;   winVal = DECAPITATES;break;
                case LIZARD:    winner = FALSE;     winVal = FALSE;     break;
                case SPOCK:     winner = PLAYER1;   winVal = SPOCK;     break;
            }
        break;
        case SPOCK:
            switch(playB){
                case ROCK:      winner = PLAYER1;   winVal = VAPORIZES; break;
                case PAPER:     winner = PLAYER2;   winVal = DISPROVES; break;
                case SCISSORS:  winner = PLAYER1;   winVal = SMASHES;   break;
                case LIZARD:    winner = PLAYER2;   winVal = POISONS;   break;
                case SPOCK:     winner = FALSE;     winVal = FALSE;     break;
            }
        break;
    }

    if (winner == PLAYER1){ //if player A wins format the strings
        sprintf(printA,"WINR%d%d%d",playA,winVal,playB);
        sprintf(printB,"LOSR%d%d%d",playA,winVal,playB);
        newGame.playerA.wins++; //increment his wins
    }
    else if (winner == PLAYER2){ //if player B wins format the strings
        sprintf(printA,"LOSR%d%d%d",playB,winVal,playA);
        sprintf(printB,"WINR%d%d%d",playB,winVal,playA);
        newGame.playerB.wins++; //increment his wins
    }
    else if (winner == FALSE){ //if it is a tie, format the strings - no winner
        sprintf(printA,"WINR%d%d%d",playA,winVal,playB);
        sprintf(printB,"WINR%d%d%d",playA,winVal,playB);
    }

    //Notify the clients of the winner/loser
    fwrite (printA , 1,  7 , newGame.playerA.pFile );
    fflush(newGame.playerA.pFile);
    fwrite (printB , 1,  7 , newGame.playerB.pFile );
    fflush(newGame.playerB.pFile);
}

/***********************************************************/
/* Function: advanceRound                                  */
/* Purpose: advances the round string and integer          */
/* Parameters: none (void)                                 */
/* Returns: none (void)                                    */
/***********************************************************/
void advanceRound(){
    sprintf(strRound,"PLAY%03d",++intRound); //advance the round string and int
}

/***********************************************************/
/* Function: do_Exit                                       */
/* Purpose: exits a server fork                            */
/* Parameters: none (void)                                 */
/* Returns: none (void)                                    */
/***********************************************************/
void do_Exit() {
    printf("The server fork is shutting down.\n");

    //close the player files
    fclose(newGame.playerA.pFile);
    fclose(newGame.playerB.pFile);

    //close the player file descriptors
    close (newGame.playerA.fileDescriptor);
    close (newGame.playerB.fileDescriptor);

    exit(0);//exit the server fork
}

/***********************************************************/
/* Function: make_server_socket                            */
/* Purpose: to start a server on the specified socket      */
/* Parameters: int portnum - the port number to use        */
/* Returns: int      the socket id                         */
/***********************************************************/
int make_server_socket(int portnum) {
    struct sockaddr_in socketAddress; //the socket address struct
    struct hostent *hostEntity; //the host struct
    char  hostname[HOSTLEN]; //the host name
    int sock_id; //the socket id

    sock_id = socket(PF_INET, SOCK_STREAM, 0); //open a socket or error
    if ( sock_id == -1 ) oops("socket");

    bzero((void *)&socketAddress, sizeof(socketAddress));//clean the sock struct

    gethostname(hostname, HOSTLEN); //get the host name
    hostEntity = gethostbyname(hostname); //then get the host

    bcopy((void *)hostEntity->h_addr, (void *)&socketAddress.sin_addr,
                    hostEntity->h_length); //copy the host info to the struct

    //build the socket info
    socketAddress.sin_port = htons(portnum);
    socketAddress.sin_family = AF_INET ;

    // bind the server to the port or error
    if ( bind(sock_id, (struct sockaddr *)&socketAddress,
            sizeof(socketAddress)) != 0 ) oops("bind");

    if ( listen(sock_id, 1) != 0 ) oops("listen");//listen on that port

    return sock_id; //return the socket id
}
